/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
        
This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/                      
#include "approximation/approxfastUtils.h"

extern int VERBOSE;

int main(int argc,char **argv) {
	if(argc < 7) {
		cout << "Usage: approximation <.mbin file> <.stac file> -k <k value> -m <m value> -v <v value> (optional, verbose)\n";
		return 1;
	}

	int K = 0, M = 0;
	int argi = 3;
	while(argi < argc) {
		if(strcmp(argv[argi], "-k") == 0)
			K = atoi(argv[argi + 1]);
		if(strcmp(argv[argi], "-m") == 0)
			M = atoi(argv[argi + 1]);
		if(strcmp(argv[argi], "-v") == 0)
			VERBOSE = atoi(argv[argi + 1]);
		argi++;
	}
	if((K == 0) || (M == 0)) {
		cout << "Usage: approximation <.mbin file> <.stac file> -k <k value> -m <m value> -v <v value> (optional, verbose)\n";
		return 1;
	}
	ifstream finmbin(argv[1], ios::in|ios::binary);
	ofstream foutstac(argv[2], ios::out);
	StdVectorFst *ifst = new StdVectorFst;
	mbintofst(ifst, finmbin);
	int numnodes = getNumNodesFst(ifst);
	int numedges = 0;
	if(numnodes < 2) {
		cout << "Error: Fst is too small!\n";
		exit(1);
	}
	//ASSUMPTION: the fst has been augmented and topologically sorted!
	//obtain adjlist of vertices and reverse too; also present nodes and base block mapping
	map <int, set<int> > adjlist;				//children on directed edges -> this global struct is the current sane whole sfa
	map <int, set<int> > adjlistrev;		//parents 
	map < Edge, block> blockdata;				//edges from the current sane whole sfa to the textual data
	//int nodepresent[numnodes];			//bitmap of whether this node is present in the full sane sfa
	for (StateIterator< Fst<StdArc> > siter(*ifst);!siter.Done(); siter.Next()) {
		int from = siter.Value();
		adjlist[from];
		adjlistrev[from];
		for(ArcIterator< Fst<StdArc> > aiter(*ifst, from);!aiter.Done(); aiter.Next()) {
			StdArc arc = aiter.Value();
			int to = arc.nextstate;
			adjlist[from].insert(to);
			adjlistrev[to].insert(from);
			string ch = ""; ch.push_back((char)arc.olabel);
			blockdata[make_pair(from,to)].pathwts.push_back(make_pair(ch, arc.weight.Value()));
			//cout << from << "\t" << to << "\t" << ch << "\t" << arc.weight.Value() << endl;
		}
	}
	int nodedismin[numnodes*numnodes];
	floydstuff(ifst, nodedismin);	//if nodedismin[i][j] is INFIN, then there is no path from i to j - this gives reachability!
	
	//sort base block mappings and crop to given K; compute the list of candidate merge pairs alongside
	set < pair <Edge, Edge> > candpairs;
	for(map<int, set<int> >::iterator miter = adjlist.begin(); miter != adjlist.end(); miter++) {
		for(set<int>::iterator siter = (miter->second).begin(); siter != (miter->second).end(); siter++) {
			numedges++;
			vector < pair<string,double> > *tmp = &(blockdata[make_pair(miter->first, *siter)].pathwts);
			sort(tmp->begin(), tmp->end(), pathwtcomp);
			if(tmp->size() > K)	//if K > num string, no need to erase anything
				tmp->erase(tmp->begin() + K, tmp->end());
			for(vector < pair<string,double> >::iterator piter = tmp->begin(); piter != tmp->end(); piter++) //cache totprob for later use
				blockdata[make_pair(miter->first, *siter)].totprob += exp(-1.0 * piter->second);
				
			/*for(set<int>::iterator niter = adjlist[*siter].begin(); niter != adjlist[*siter].end(); niter++) //if this node has children, add successive edge pairs
				if(candpairs.find(make_pair(make_pair(*siter,*niter),make_pair(miter->first,*siter))) == candpairs.end())	//rev pair is not there
					candpairs.insert(make_pair(make_pair(miter->first,*siter),make_pair(*siter,*niter)));
			*/
			//instead of above (add outdeg child pairs to cands), we need add just 1 pair with a child!! because, findminsfa result will be identical for all such pairs!!
			//this could result in very significant savings in computations of findminsfa! in fact, it reduces by a third or so!
			if(adjlist[*siter].size() > 0) {
				int achild = *(adjlist[*siter].begin());
				if(candpairs.find(make_pair(make_pair(*siter,achild),make_pair(miter->first,*siter))) == candpairs.end())	//rev pair is not there
					candpairs.insert(make_pair(make_pair(miter->first,*siter),make_pair(*siter,achild)));	
			}
		}

		for(set<int>::iterator siter1 = (miter->second).begin(); siter1 != (miter->second).end(); siter1++) 		//deg c 2 outgoing sibling edge pairs
			for(set<int>::iterator siter2 = (miter->second).begin(); siter2 != (miter->second).end(); siter2++)
					if((*siter1 != *siter2) && (candpairs.find(make_pair(make_pair(miter->first,*siter2),make_pair(miter->first,*siter1))) == candpairs.end()))
						candpairs.insert(make_pair(make_pair(miter->first,*siter1),make_pair(miter->first,*siter2)));

		for(set<int>::iterator siter1 = adjlistrev[miter->first].begin(); siter1 != adjlistrev[miter->first].end(); siter1++) 		//deg c 2 _incoming_ sibling edge pairs
			for(set<int>::iterator siter2 = adjlistrev[miter->first].begin(); siter2 != adjlistrev[miter->first].end(); siter2++)
					if((*siter1 != *siter2) && (candpairs.find(make_pair(make_pair(*siter2,miter->first),make_pair(*siter1,miter->first))) == candpairs.end()))
						candpairs.insert(make_pair(make_pair(*siter1,miter->first),make_pair(*siter2,miter->first)));
	}
	//cout << "k " << K << " m " << M << " num blockdata: " << blockdata.size() << endl;
	cout << "num candpairs: " << candpairs.size() << endl;
	
	//compute overall prob now
	map <int, double> memoprob;
	double globaloverallprob = probDynMult(&adjlist, &blockdata, 0, &memoprob);
	cout << "global overallprob: " << globaloverallprob << endl;

	if(blockdata.size() <= M) { //no need for merges
		if(VERBOSE >= 1) cout << "num blks " << blockdata.size() << " <= M " << M << "; no need to merge\n";
	}
	else {
	//iterate through the list of cand merge pairs, for each: do findminsfa, obtain set of edges present in the resultant sfa - obtain the final set of cand mergedsfas
	//a cand mergedsfa (subsfa) is simply a set of edges (generlzn of edge pairs)
	set <blockskel> candsfas;
	int cntr=1;
	for(set < pair <Edge, Edge> >::iterator iter = candpairs.begin(); iter != candpairs.end(); iter++) {
if(VERBOSE >= 1)	cout << "***cntr " << cntr++ << " calling findminsfa on pair: (" << (*iter).first.first << "," << (*iter).first.second << ")-(" << (*iter).second.first << "," << (*iter).second.second << "):\t";

		blockskel tmp = findMinSfa(&adjlist, &adjlistrev, nodedismin, numnodes, *iter);
		candsfas.insert(tmp);

if(VERBOSE >= 1) cout << " (" << tmp.minnode << "," << tmp.maxnode << ")\n";

	}//end for on cand pairs
	
	cout << "num cand merged minsfas: " << candsfas.size() << endl;
	
	if(VERBOSE >= 1) {
		cout << "cand merged minsfa are: min-max, numnodes, numedges (and subsfa)\n";
		for(set <blockskel>::iterator iter = candsfas.begin(); iter != candsfas.end(); iter++) {
			cout << iter->minnode << "-" << iter->maxnode << ", " << (iter->sfa).size() << ", ";
			map <int, set <int> > tmp = iter->sfa;					int numedges = 0;
			for(map <int, set <int> >::iterator ssiter = tmp.begin(); ssiter != tmp.end(); ssiter++)
				numedges += (ssiter->second).size();
			cout << numedges;
			if(VERBOSE == 1) cout << endl;
			if(VERBOSE >= 2) {
				cout << " and subsfa:\n";
				for(map <int, set <int> >::iterator ssiter = tmp.begin(); ssiter != tmp.end(); ssiter++) {
					cout << "\t" << ssiter->first << ":";
					for(set <int>::iterator essiter = (ssiter->second).begin(); essiter != (ssiter->second).end(); essiter++)
						cout << " " << *essiter;
					cout << endl;
				}
			}
		}
	}
	
	cout << "computing cand overallprobs:\n";
	map < map <int, set <int> >, double> candsfaprobs;		//maps from cached cand sfa to its overallprob
	
	//now, iterate on cand mergedsfas: temprly update overall data structs incl blkdata, compute overallprob and memoize; get best overallprob one for next step
	double maxprob = 0.0;										//maximum overallprob among the set of integrated mergedsfas 
	map <int, set<int> > maxadjlist;				//children on directed edges -> this global struct is the current sane whole sfa
	map <int, set<int> > maxadjlistrev;			//parents 
	map < Edge, block> maxblockdata;				//edges from the current sane whole sfa to the textual data	
	blockskel chosenskel;
	
	int scratch[numnodes];									//used as bitmap for subfst projections
	for(set <blockskel>::iterator iter = candsfas.begin(); iter != candsfas.end(); iter++) {
		memset(scratch, '\0', sizeof(int) * numnodes);	//bitmap is cleared for every projection 
		if(VERBOSE >= 1) cout << "getblkdata for skel: " << (*iter).minnode << "-" << (*iter).maxnode << " n " << ((*iter).sfa).size() << endl;
		block thisblock = getBlockData(*iter, &blockdata, K, 1, scratch);
		//block thisblock = getBlockData(ifst, *iter, numnodes, K, 1, scratch);	//1 means, we store paths/wts; we dont use blkdata/adjlist coz topk uses chars, but it is correct
		if(VERBOSE >= 1) cout << "\ttotprob is " << thisblock.totprob << " and firstpathwt: " << thisblock.pathwts[0].first << "\t" << thisblock.pathwts[0].second << endl;
		
		map <int, set<int> > tmpadjlist = adjlist; //copy - this seems slightly faster than the updatenocopy
		map <int, set<int> > tmpadjlistrev = adjlistrev;
		map <Edge, block> tmpblockdata = blockdata;
		updateStructures(*iter, &thisblock, &tmpadjlist, &tmpadjlistrev, &tmpblockdata);	//tmply update global structs to compute overallprob
		
		map <int, double> tmpmemo;
		double tmpprob = probDynMult(&tmpadjlist, &tmpblockdata, 0, &tmpmemo);
		candsfaprobs[(*iter).sfa] = tmpprob;		//cache the cand overallprobs
		if(VERBOSE >= 1) cout << "\tcached overallprob for this skel is " << tmpprob << endl;
		
		if(tmpprob > maxprob) {		//get max
			maxprob = tmpprob;
			maxadjlist = tmpadjlist;
			maxadjlistrev = tmpadjlistrev;
			maxblockdata = tmpblockdata;
			chosenskel = *iter;
		}
	}//end for on prob comptn on candsfas

	adjlist = maxadjlist;
	adjlistrev = maxadjlistrev;
	blockdata = maxblockdata;
	globaloverallprob = maxprob;

	if(VERBOSE >= 0) cout << "after inital itern on cands, chosenskel " << chosenskel.minnode << "-" << chosenskel.maxnode << "n" << chosenskel.sfa.size() << " overallprob " << globaloverallprob << " adjlist.size " << adjlist.size() << " numblks " << blockdata.size() << " while M " << M << endl;

	//start loop over the remanining dep of m; basically, three steps:
	//1 remove the chosen cand from memos
	//2 remove candts in memo have a partial overlap with chosen cand (if candt >= chosen cand, ie is a superset or equal, retain them - NO!! since i am not using origfst anymore, i need to modify the adjlist of this superset cand, otherwise, later during blk construcn, it will be a bug since some edges wont be in then adjlist)
	//3 get the ngbrs of chosen cand, and the corresp set of legal cand skels; unite with memoized set
	while (blockdata.size() > M) {
		
		if(VERBOSE >= 0) cout << "in cycle, numblks " << blockdata.size() << " > M " << M << endl;
		
		candsfas.erase(chosenskel);	//step 1
		candsfaprobs.erase(chosenskel.sfa);
		
		if(VERBOSE >= 1) cout << "\t done step 1 of erasing chosenskel " << chosenskel.minnode << "-" << chosenskel.maxnode << "n" << chosenskel.sfa.size() << " and candsfas size " << candsfas.size() << endl;
				
		set < Edge > chosenskeledges = getEdgeSet(chosenskel.sfa); //step 2
		vector <blockskel> tobedeleted;		//list of keys into memos to be deleted due to partial overlap
		vector < pair <blockskel, double> > tobereplaced;		//skels to be reinserted (post deletion above) after adjlists adjustments
		for(set <blockskel>::iterator miter = candsfas.begin(); miter != candsfas.end(); miter++) {
			//check the intersection; if it is not null and not equal to the chosen one, it means there is a partial overlaps and we need to delete it
			set < Edge > intsxn = intersectEdgeSets(chosenskeledges, getEdgeSet((*miter).sfa));
			if (intsxn == chosenskeledges) {	//cand is superset of chosenskel; so modify the cand's adjlists to reflect the changes imposed by collapse of chosenskel
				blockskel tmp = *miter;
				double tmppro = candsfaprobs[(*miter).sfa];
				for(map <int, set <int> >::iterator cciter = (chosenskel.sfa).begin(); cciter != (chosenskel.sfa).end(); cciter++) {
					for(set <int>::iterator csiter = (cciter->second).begin(); csiter != (cciter->second).end(); csiter++) {	//drop all the edges of the skel from the cand
						tmp.sfa[cciter->first].erase(*csiter);									tmp.sfarev[*csiter].erase(cciter->first);
					}
					if((cciter->first != chosenskel.minnode) && (cciter->first != chosenskel.maxnode)) {	//drop all interior nodes of the skel from the cand
						(tmp.sfa).erase(cciter->first);								(tmp.sfarev).erase(cciter->first);
					}
					tmp.sfa[chosenskel.minnode].insert(chosenskel.maxnode);				tmp.sfarev[chosenskel.maxnode].insert(chosenskel.minnode);	//collapsed edge
				}
				tobereplaced.push_back(make_pair(tmp, tmppro));
				tobedeleted.push_back(*miter);	//new one will be reinserted later 				
			}
			else if (intsxn.size() > 0)		//not superset but only a partial overlap
				tobedeleted.push_back(*miter);
		}
		if(VERBOSE >= 1) cout << "\tin step 2, num partial overlappers to delete: " << tobedeleted.size() << endl;
		for(vector <blockskel>::iterator diter = tobedeleted.begin(); diter != tobedeleted.end(); diter++) {
			candsfas.erase(*diter);
			candsfaprobs.erase((*diter).sfa);
		}
		if(VERBOSE >= 1) cout << "\tin step 2, num superior overlappers to reinsert: " << tobereplaced.size() << endl;
		for(vector < pair<blockskel, double> >::iterator rditer = tobereplaced.begin(); rditer != tobereplaced.end(); rditer++) {
			candsfas.insert(rditer->first);
			candsfaprobs[(rditer->first).sfa] = rditer->second;
		}
		if(VERBOSE >= 1) cout << "\tdone step 2 of erasing partial overlappers and replacing superiors" << endl;
		
		//step 3
		//with the new collapsed edge, we can get many new candpairs: seq with any 1 prev of min, seq with any 1 next of max, n siblings out min and in max
		set < pair <Edge, Edge> > morecandpairs;
		Edge collapsedge = make_pair(chosenskel.minnode, chosenskel.maxnode);
		
		if(adjlistrev[chosenskel.minnode].size() > 0)
			morecandpairs.insert(make_pair(collapsedge, make_pair(*(adjlistrev[chosenskel.minnode].begin()), chosenskel.minnode)));
		if(adjlist[chosenskel.maxnode].size() > 0)
			morecandpairs.insert(make_pair(collapsedge, make_pair(chosenskel.maxnode, *(adjlist[chosenskel.maxnode].begin()))));

		if(VERBOSE >= 1) cout << "\tin step 3, added 2 new seq pairs morecandpairs size " << morecandpairs.size() << endl;
		
		for(set<int>::iterator siter = adjlist[chosenskel.minnode].begin(); siter != adjlist[chosenskel.minnode].end(); siter++) 		//deg-1 outgoing sibling edge pairs
			if(*siter != chosenskel.maxnode)	//dont repeat collapsed edge
				morecandpairs.insert(make_pair(collapsedge, make_pair(chosenskel.minnode, *siter)));
		
		if(VERBOSE >= 1) cout << "\tin step 3, added outgoing sibling edgepairs morecandpairs size " << morecandpairs.size() << endl;
				
		for(set<int>::iterator siter = adjlistrev[chosenskel.maxnode].begin(); siter != adjlistrev[chosenskel.maxnode].end(); siter++) //deg-1 incoming sibling edge pairs
			if(*siter != chosenskel.minnode)	//dont repeat collapsed edge
				morecandpairs.insert(make_pair(collapsedge, make_pair(*siter, chosenskel.maxnode)));
		
		if(VERBOSE >= 1) {
		cout << "\tin step 3 morecandpairs size " << morecandpairs.size() << " and they are: ";
		for(set < pair <Edge, Edge> >::iterator mciter =  morecandpairs.begin(); mciter != morecandpairs.end(); mciter++) {
			cout << "(" << mciter->first.first << "-" << mciter->first.second << ")-(" << mciter->second.first << "-" << mciter->second.second << ") ; ";
		}
		cout << endl;
		}
		
		//findminsfa on the morecandpairs to get more candsfas to add into candsfas

		for(set < pair <Edge, Edge> >::iterator iter = morecandpairs.begin(); iter != morecandpairs.end(); iter++) {

if(VERBOSE >= 1)	cout << "***cntr " << cntr++ << " calling findminsfa on morepair: (" << (*iter).first.first << "," << (*iter).first.second << ")-(" << (*iter).second.first << "," << (*iter).second.second << "):\t";

			blockskel tmp = findMinSfa(&adjlist, &adjlistrev, nodedismin, numnodes, *iter);
			candsfas.insert(tmp);

if(VERBOSE >= 1) cout << " (" << tmp.minnode << "," << tmp.maxnode << ")\n";

		}//end for on morecand pairs
	
		if(VERBOSE >= 1) cout << "num cand merged minsfas after augmenting: " << candsfas.size() << endl;
	
		if(VERBOSE >= 1) {
			cout << "cand merged minsfa are: min-max, numnodes, numedges (and subsfa)\n";
			for(set <blockskel>::iterator iter = candsfas.begin(); iter != candsfas.end(); iter++) {
				cout << iter->minnode << "-" << iter->maxnode << ", " << (iter->sfa).size() << ", ";
				map <int, set <int> > tmp = iter->sfa;					int numedges = 0;
				for(map <int, set <int> >::iterator ssiter = tmp.begin(); ssiter != tmp.end(); ssiter++)
					numedges += (ssiter->second).size();
				cout << numedges;
				if(VERBOSE == 1) cout << endl;
				if(VERBOSE >= 2) {
					cout << " and subsfa:\n";
					for(map <int, set <int> >::iterator ssiter = tmp.begin(); ssiter != tmp.end(); ssiter++) {
						cout << "\t" << ssiter->first << ":";
						for(set <int>::iterator essiter = (ssiter->second).begin(); essiter != (ssiter->second).end(); essiter++)
							cout << " " << *essiter;
						cout << endl;
					}
				}
			}
		}
	
		if(VERBOSE >= 1) cout << "computing new cand overallprobs for candsfas.size :" << candsfas.size() << endl;
		//map < map <int, set <int> >, double> candsfaprobs;		//maps from cached cand sfa to its overallprob
	
		//now, iterate on cand mergedsfas: temprly update overall data structs incl blkdata, compute overallprob and memoize; get best overallprob one for next step
		double newmaxprob = 0.0;										//maximum overallprob among the set of integrated mergedsfas 
		blockskel newchosenskel;
		
		int newscratch[numnodes];									//used as bitmap for subfst projections
		for(set <blockskel>::iterator iter = candsfas.begin(); iter != candsfas.end(); iter++) {
			//check if blk exists in memo; if so, skip getting blk again; but we still need to do update structs on the maxprob skel to get adjlist, etc.
			map < map <int, set <int> >, double >::iterator found = candsfaprobs.find((*iter).sfa);
			if(found != candsfaprobs.end()) {
				if((found->second) > newmaxprob) {
					newmaxprob = found->second;
					newchosenskel = *iter;
				}
			}
			else {				//this is a new cand mergedsubsfa since prob not in cache; need to recompute
				memset(scratch, '\0', sizeof(int) * numnodes);	//bitmap is cleared for every projection 
if(VERBOSE >= 1) cout << "getblkdata for new skel: " << (*iter).minnode << "-" << (*iter).maxnode << " n " << ((*iter).sfa).size() << endl;
				block thisblock = getBlockData(*iter, &blockdata, K, 1, scratch);
				//block thisblock = getBlockData(ifst, *iter, numnodes, K, 1, scratch);
if(VERBOSE >= 1) cout << "\ttotprob is " << thisblock.totprob << " and firstpathwt: " << thisblock.pathwts[0].first << "\t" << thisblock.pathwts[0].second << endl;
				map <int, set<int> > tmpadjlist = adjlist;
				map <int, set<int> > tmpadjlistrev = adjlistrev;
				map <Edge, block> tmpblockdata = blockdata;
				updateStructures(*iter, &thisblock, &tmpadjlist, &tmpadjlistrev, &tmpblockdata);
				map <int, double> tmpmemo;
				double tmpprob = probDynMult(&tmpadjlist, &tmpblockdata, 0, &tmpmemo);
				candsfaprobs[(*iter).sfa] = tmpprob;		//cache the new cand overallprobs
if(VERBOSE >= 1) cout << "\tcached overallprob for this new skel is " << tmpprob << endl;
				if(tmpprob > newmaxprob) {		//get max
					newmaxprob = tmpprob;
					newchosenskel = *iter;
				}
			}//end else on if prob is in cache
		}//end for on prob comptn on candsfas

		//now directly update the global overall structs with the maxblk!
		memset(scratch, '\0', sizeof(int) * numnodes);
		block maxblock = getBlockData(newchosenskel, &blockdata, K, 1, scratch);
		//block maxblock = getBlockData(ifst, newchosenskel, numnodes, K, 1, scratch);
		updateStructures(newchosenskel, &maxblock, &adjlist, &adjlistrev, &blockdata);
		globaloverallprob = newmaxprob;
		chosenskel = newchosenskel;

if(VERBOSE >= 0)	cout << "after cycle on new cands, newchosenskel " << newchosenskel.minnode << "-" << newchosenskel.maxnode << "n" << newchosenskel.sfa.size() << " overallprob " << globaloverallprob << " new adjlist.size " << adjlist.size() << " new numblks " << blockdata.size() << endl;

	}//end while on blkdata > m
	}//end if num blks < m
	cout << "finally, num blks: " << blockdata.size() << " adjlist size " << adjlist.size() << " globaloverallprob " << globaloverallprob << endl;
	
	//edge push to get blk graph and mapping and print to file
	//stacfile schema:
	//<M> <K> <number of edges in fst> <numnodes in fst>\n
	//<final numblks> <overall prob of final blocks set>\n
	foutstac << M << " " << K << " " << numedges << " " << numnodes << endl;
	foutstac << blockdata.size() << " " << globaloverallprob << endl;
	
	map <Edge, int> blocknumbers;				//the block/edge numbering to be used for stacgraph
	int startnum = 0;
	for(map <int, set <int> >::iterator iter = adjlist.begin(); iter != adjlist.end(); iter++)
		for(set <int>::iterator siter = (iter->second).begin(); siter != (iter->second).end(); siter++)
			blocknumbers[make_pair(iter->first,*siter)] = startnum++;
			
	//print out the stacgraph - look ahead two edges, and use the edge/block numbers to output the stac blk gph directly
	for(map <int, set <int> >::iterator iter = adjlist.begin(); iter != adjlist.end(); iter++) {
		for(set <int>::iterator siter = (iter->second).begin(); siter != (iter->second).end(); siter++) {		//for first edge
			foutstac << blocknumbers[make_pair(iter->first,*siter)] << "\t";
			for(set <int>::iterator nsiter = adjlist[*siter].begin(); nsiter != adjlist[*siter].end(); nsiter++)	//for second edges
				foutstac << blocknumbers[make_pair(*siter,*nsiter)] << "\t";
			foutstac << endl;
		}
	}

	for(map <Edge, block>::iterator biter = blockdata.begin(); biter != blockdata.end(); biter++) {
		foutstac << blocknumbers[biter->first] << " 1 " << biter->second.pathwts.size() << endl;	//1 (prevly num origedges is never used, kept for stac format only)
		for(vector < pair<string,double> >::iterator paiter = biter->second.pathwts.begin(); paiter != biter->second.pathwts.end(); paiter++)
			foutstac << paiter->first << "\n" << paiter->second << endl;
	}
	
	return 0;
}
