/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "libpq-fe.h"
#include "libpq/libpq-fs.h"
#include "frontends/SQLUtils.h"
#include "libraries/fstUtils.h"
#include "libraries/regexParser.h"

int main(int argc,char **argv) {
	if(argc < 5) {
		printUsage(FULLFST);
		return 1;
	}
	arguments theargs;
	theargs.populate(argc, argv);
	theargs.check(FULLFST);
	querystructs qstructs = parseQuery((char*)theargs.sqlquery.c_str());
	StdVectorFst query;
	int qr = getQueryFst(&query, (char*)(theargs.regex).c_str());
	if(qr == -1) {
		cout << "Error: Malformed regex!\n";
		return 1;
	}
	if(theargs.V) cout << "Parsed query\n";
	
	PGconn* conn;
	string connval = "dbname=" + theargs.dbname + " port=" + theargs.portnum;
	conn = PQconnectdb(connval.c_str());
	if(PQstatus(conn) != CONNECTION_OK) {
		cout << "Error: Could not connect with " << connval << "! Message: " << PQerrorMessage(conn);
		PQfinish(conn);
		exit(1);
	}
	if(theargs.V) cout << "Connected to database\n";	
	
	PGresult* res;
	int numac = qstructs.auxcolnames.size();
	string cmdstring = qstructs.constructCmd();
	res = PQexec(conn, (char*) cmdstring.c_str());
	if(PQresultStatus(res) != PGRES_TUPLES_OK) {
		cout << PQerrorMessage(conn);
		PQclear(res);
		PQfinish(conn);
		exit(1);
	}

	int numt = PQntuples(res);
	if(theargs.V) cout << "Reading " << numt << " tuples\n";
	vector <restuple> results;
	string cmdstringblob = "select fstblob from " + qstructs.tablename + "_fullfst where dataid = $1;";
	PQprepare(conn, "getfst", (const char*)cmdstringblob.c_str(), 1, NULL);
	PQprepare(conn, "begintr", "begin;", 0, NULL);
	PQprepare(conn, "endtr", "end;", 0, NULL);

	for(int i=0; i <numt; i++) {
		string thiskey = PQgetvalue(res, i, 0);
		string thisid = PQgetvalue(res, i, 1);	
		if(theargs.V) cout << "Computing for tuple dataid " << thisid << " datakey " << thiskey << endl;;
		PGresult *resblob;
		const char *paramValuesblob[1];
		paramValuesblob[0] = (const char*) thisid.c_str();	
		resblob = PQexecPrepared(conn, "getfst", 1, paramValuesblob, NULL, NULL, 1);

		PQexecPrepared(conn, "begintr", 0, NULL, NULL, NULL, 1);		
		Oid objid = unpack_int32(PQgetvalue(resblob, 0, 0)); 
		int objfd = lo_open(conn, objid, INV_READ);
		lo_lseek(conn, objfd, 0, SEEK_END);
		int bsize = lo_tell(conn, objfd);
		lo_lseek(conn, objfd, 0, SEEK_SET);
		char buffer[bsize];
		memset(buffer, bsize, '\0');		
		int retlo = lo_read(conn, objfd, buffer, bsize);
		lo_close(conn, objfd);
		PQexecPrepared(conn, "endtr", 0, NULL, NULL, NULL, 1);
		if(retlo < 0) {
			if(theargs.V) cout << "Warning: Failed to read fstblob at dataid " << thisid << " datakey " << thiskey << endl;
			continue;
		}
		
		StdVectorFst *ifst = new StdVectorFst;
		mbintofst(ifst, buffer);	//deserialization of mbin to in-memory fst
		vector<double> dynreachprobs = dynReachProbs(ifst);	//reaching probs
		addEpsFst(ifst);
		StdVectorFst unionnc2 (ComposeFst<StdArc>::ComposeFst(*ifst, query));
		Connect(&unionnc2);		//this will remove arcs if no +prob match exist
		delete ifst;

		int numunionnodes = getNumNodesFst(&unionnc2);
		if(numunionnodes < 2) { //no answers at all; go to next fst
			if(theargs.V) cout << "Warning: No answers from fstblob for dataid " << thisid << " datakey " << thiskey << endl;
			continue;
		};
		PathIterator incrementalN;
		constructIterator(&incrementalN, &unionnc2);
		Path p = incrementalN.getNext();
		restuple thistuple;
		double candwt = p.getPathCost();	//we consider only top path prob!
		thistuple.matchwt = candwt - log(dynreachprobs[p.getEdge(1)->origfrom]); //mult by prob of reaching the start node in this nodeseq
		for(int colv=0; colv<numac; colv++)
			thistuple.precolvals.push_back(PQgetvalue(res, i, 2+colv));	//2 for datakey and dataid, and colvs for auxcols
		results.push_back(thistuple);
		delete incrementalN.enumerator;
	}//end for on fsts
	
	PQclear(res);
	PQfinish(conn);

	sortResults(&results);
	printResults(results, qstructs.auxcolnames, qstructs.Ktop);

return 0;
}
