/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "libpq-fe.h"
#include "libpq/libpq-fs.h"
#include "frontends/SQLUtils.h"
#include "libraries/fstUtils.h"
#include "libraries/regexParser.h"
#include "frontends/staccatoSQLUtils.h"

int main(int argc,char **argv) {
	if(argc < 9) {
		printUsage(STAC);
		return 1;
	}
	arguments theargs;
	theargs.populate(argc, argv);
	theargs.check(STAC);
	querystructs qstructs = parseQuery((char*)theargs.sqlquery.c_str());
	string starregex = string("(\\x)*") + theargs.regex;	//the query match can start anywhere in the fst
	StdVectorFst tquery;
	int qr = getQueryFst(&tquery, (char*)starregex.c_str());
	if(qr == -1) {
		cout << "Error: Malformed regex!\n";
		return 1;
	}
	StdVectorFst query;
	Determinize(tquery, &query);
	Connect(&query);
	TopSort(&query);	//doesnt affect if query has loops; in case of kleene star, dfa does have loops
	int startstate = 0;
	int finalstate = getFinalStateFst(&query);
	int NUMST = getNumNodesFst(&query);
	int dfaquery [NUMST * 94];		//given the statenum and the char read, we get the (single) next state to go to
	getQueryMatrix(&query, dfaquery);
	if(theargs.V) cout << "Parsed query\n";
	
	PGconn* conn;	
	string connval = "dbname=" + theargs.dbname + " port=" + theargs.portnum;
	conn = PQconnectdb(connval.c_str());
	if(PQstatus(conn) != CONNECTION_OK) {
		cout << "Error: Could not connect with " << connval << "! Message: " << PQerrorMessage(conn);
		PQfinish(conn);
		exit(1);
	}
	if(theargs.V) cout << "Connected to database\n";
	
	PGresult* res;
	int numac = qstructs.auxcolnames.size();
	string cmdstring = qstructs.constructCmd();
	res = PQexec(conn, (char*) cmdstring.c_str());
	if(PQresultStatus(res) != PGRES_TUPLES_OK) {
		cout << PQerrorMessage(conn);
		PQclear(res);
		PQfinish(conn);
		exit(1);
	}

	int numt = PQntuples(res);
	if(theargs.V) cout << "Reading " << numt << " tuples\n";
	vector <restuple> results;	
	stringstream Mstr, Kstr;
	Mstr << theargs.M;
	Kstr << theargs.K;
	string stacgmtablename = qstructs.tablename + "_staccato_" + Mstr.str() + "_" + Kstr.str();
	string stacgmblkgphtablename = stacgmtablename + "_graph";
	string cmdstringpaths2 = "select blocknum, path, pathwt from " + stacgmtablename + " where dataid = $1;";
	string cmdstringgph = "select graph from " + stacgmblkgphtablename + " where dataid = $1;";
	
	PQprepare(conn, "getpaths", (const char*)cmdstringpaths2.c_str(), 1, NULL);
	PQprepare(conn, "getgph", (const char*)cmdstringgph.c_str(), 1, NULL);
	PQprepare(conn, "begintr", "begin;", 0, NULL);
	PQprepare(conn, "endtr", "end;", 0, NULL);
	
	for(int ti=0; ti <numt; ti++) {		//iterate over each fst
		string thiskey = PQgetvalue(res, ti, 0);
		string thisid = PQgetvalue(res, ti, 1);
		if(theargs.V) cout << "Computing for tuple dataid " << thisid << " datakey " << thiskey << endl;;
		PGresult* resgph;
		const char *paramValuesgph[1];
		paramValuesgph[0] = (const char*) thisid.c_str();
		resgph = PQexecPrepared(conn, "getgph", 1, paramValuesgph, NULL, NULL, 1);
		PQexecPrepared(conn, "begintr", 0, NULL, NULL, NULL, 1);
		Oid objid = unpack_int32(PQgetvalue(resgph, 0, 0));
		int objfd = lo_open(conn, objid, INV_READ);
		lo_lseek(conn, objfd, 0, SEEK_END);
		int bsize = lo_tell(conn, objfd);
		lo_lseek(conn, objfd, 0, SEEK_SET);
		char buffer[bsize];
		memset(buffer, bsize, '\0');
		int retlo = lo_read(conn, objfd, buffer, bsize);
		if(retlo < 0) {
			if(theargs.V) cout << "Warning: Failed to read graph blob at dataid " << thisid << endl;
			continue;
		}
		lo_close(conn, objfd);
		PQexecPrepared(conn, "endtr", 0, NULL, NULL, NULL, 1);
		
		structureInfo stacinfo;
		graphread(&stacinfo, buffer);
		PQclear(resgph);

		PGresult* resti;
		const char *paramValues[1];
		paramValues[0] = (const char*) thisid.c_str();
		resti = PQexecPrepared(conn, "getpaths", 1, paramValues, NULL, NULL, 1);
		if (PQresultStatus(resti) != PGRES_TUPLES_OK) {
			cout << "Error: Select tuples failed: " << PQerrorMessage(conn);
			PQclear(resti);
			PQfinish(conn);
			return 1;
		}
		int numtti = PQntuples(resti);
		if(numtti <= 0) {
			if(theargs.V) cout << "Warning: No tuples were read for dataid " << thisid << "\n";
			continue;
		}
	
		int numblks = stacinfo.stacgraph.size();
		double* transprob = (double *) malloc (sizeof(double) * numblks * NUMST * NUMST);	//allocate array of matrices for each block in gph
		memset(transprob, '\0', sizeof(double) * numblks * NUMST * NUMST);
		for(int x=0; x<numtti; x++) {
			char* bnox = PQgetvalue(resti, x, 0);
			char* bpath = PQgetvalue(resti, x, 1);
			char* bwt = PQgetvalue(resti, x, 2);
			int nm = unpack_int32(bnox);
			string pathx = string(bpath);
			double wt = unpack_double(bwt);
			for(int inp=0; inp<NUMST; inp++) {
				int linenxtst = dfacheck(pathx, dfaquery, inp , NUMST, startstate, finalstate);
				transprob [nm * NUMST * NUMST + inp * NUMST + linenxtst] += exp(-1.0*wt);
			}
		}
		PQclear(resti);
		//the stacgraph may not have a unique start node; since, we could have branching at 0 in the fst, mult blks could hv indeg 0 in the stacgraph
		//so, compute the matrmultdyn for all start node, and simply add them up; we can also memoize the blknums's retval seen before in the function
		int* memo = (int *) malloc (sizeof(int) * numblks);
		memset(memo, '\0', sizeof(int) * numblks);
		map <int, vector <int> > stacgraphrev;
		reverseGraph(&stacinfo.stacgraph, &stacgraphrev);
		vector <int> startblks;
		for(map <int, vector <int> >::iterator iter = stacgraphrev.begin(); iter != stacgraphrev.end(); iter++)
			if(iter->second.size() == 0)	//indeg is 0 means it is a startblknode
				startblks.push_back(iter->first);
		double* probmatr = (double *) malloc (sizeof(double) * NUMST * NUMST);	//allocate array of matrices for each block in gph
		memset(probmatr, '\0', sizeof(double) * NUMST * NUMST);
		double* scratches = (double *) malloc (sizeof(double) * 2 * NUMST * NUMST);	//scratches for addn and mult of matrices; memset done in function
		for(vector <int>::iterator siter = startblks.begin(); siter != startblks.end(); siter++) {
			//compute this statrnode's prob matr, and add to the final result
			matrdfs (transprob, scratches, memo, NUMST, &stacinfo.stacgraph, *siter);
			incrsummatrs (probmatr, transprob + ((*siter) * NUMST * NUMST), NUMST);
		}

		double acceptprob = probmatr[startstate * NUMST + finalstate];
		free(transprob);
		free(memo);
		free(scratches);
		free(probmatr);
		restuple thistuple;
		if(acceptprob > 0) { //only consider those with matchprob > 0
			thistuple.matchwt = -1.0*log(acceptprob);
			for(int colv=0; colv<numac; colv++)
				thistuple.precolvals.push_back(PQgetvalue(res, ti, 2+colv));	//auxcols were obtained with the keys in fst itern
			results.push_back(thistuple);
		}
	}//end for on fsts
	
	PQclear(res);
	PQfinish(conn);
	
	sortResults(&results);
	printResults(results, qstructs.auxcolnames, qstructs.Ktop);

return 0;
}
