/*
Copyright 2012 Arun Kumar and Christopher Re

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.

This work was supported by the Microsoft Jim Gray Systems Lab, 
the National Science Foundation and the Office of Naval Research.
*/
#include "libpq-fe.h"
#include "libpq/libpq-fs.h"
#include "frontends/SQLUtils.h"
#include "libraries/fstUtils.h"
#include "libraries/regexParser.h"
#include "frontends/staccatoSQLUtils.h"
#include "frontends/staccatoSQLIndexUtils.h"

int main(int argc,char **argv) {
	if(argc < 11) {
		printUsage(STACPROJ);
		return 1;
	}
	arguments theargs;
	theargs.populate(argc, argv);
	theargs.check(STACPROJ);
	querystructs qstructs = parseQuery((char*)theargs.sqlquery.c_str());
	string starregex = string("(\\x)*") + theargs.regex; //the query match can start anywhere in the fst
	StdVectorFst tquery;
	int qr = getQueryFst(&tquery, (char*)starregex.c_str());
	if(qr == -1) {
		cout << "Error: Malformed regex!\n";
		return 1;
	}
	StdVectorFst query;
	Determinize(tquery, &query);
	Connect(&query);
	TopSort(&query);	//doesnt affect if query has loops; in case of kleene star, dfa does have loops
	int startstate = 0;
	int finalstate = getFinalStateFst(&query);
	int NUMST = getNumNodesFst(&query);
	int dfaquery [NUMST * 94];		//given the statenum and the char read, we get the (single) next state to go to
	getQueryMatrix(&query, dfaquery);
	if(theargs.V) cout << "Parsed query\n";
	
	PGconn* conn;	
	string connval = "dbname=" + theargs.dbname + " port=" + theargs.portnum;
	conn = PQconnectdb(connval.c_str());
	if(PQstatus(conn) != CONNECTION_OK) {
		cout << "Error: Could not connect with " << connval << "! Message: " << PQerrorMessage(conn);
		PQfinish(conn);
		exit(1);
	}
	if(theargs.V) cout << "Connected to database\n";

	//we first retrieve the list of fsts and postings where the term occurs from the index
	PGresult* res;
	stringstream Mstr, Kstr;
	Mstr << theargs.M;
	Kstr << theargs.K;
	string stacgmtablename = qstructs.tablename + "_staccato_" + Mstr.str() + "_" + Kstr.str();
	string stacgmblkgphtablename = stacgmtablename + "_graph";
	int numac = qstructs.auxcolnames.size();	
	string cmdstring = "select dataid, blocknum, pathnum, offsetnum from " + stacgmtablename + "_index where term = '" + theargs.indexterm + "';";	//only index access
	res = PQexecParams(conn, (char*) cmdstring.c_str(), 0, NULL, NULL, NULL, NULL, 1);
	if(PQresultStatus(res) != PGRES_TUPLES_OK) {
		cout << PQerrorMessage(conn);
		PQclear(res);
		PQfinish(conn);
		exit(1);
	}
	int numall = PQntuples(res);
	map <string, vector< pair <int, pair <int, int> > > > fstpostings;
	for(int y=0; y<numall; y++) {
		int thisidint = unpack_int32(PQgetvalue(res, y, 0));
		stringstream thisidstr;
		thisidstr << thisidint;
		string thisid = thisidstr.str();
		int blkno, pathno, offsetno;
		blkno = unpack_int32(PQgetvalue(res, y, 1));
		pathno = unpack_int32(PQgetvalue(res, y, 2));
		offsetno = unpack_int32(PQgetvalue(res, y, 3));
		fstpostings[thisid].push_back(make_pair(blkno, make_pair(pathno, offsetno)));	//since there could be mult postings per fst
	}

	int numt = fstpostings.size();	//number of uniq fsts
	if(theargs.V) cout << "Obtained " << numall << " postings with " << numt << " fsts from index\n";
	vector <restuple> results;
	string cmdstringpaths2 = "select blocknum, path, pathwt from " + stacgmtablename + " where dataid = $1;";
	string cmdstringgph = "select graph from " + stacgmblkgphtablename + " where dataid = $1;";
	PQprepare(conn, "getpaths", (const char*)cmdstringpaths2.c_str(), 1, NULL);
	PQprepare(conn, "getgph", (const char*)cmdstringgph.c_str(), 1, NULL);
	PQprepare(conn, "begintr", "begin;", 0, NULL);
	PQprepare(conn, "endtr", "end;", 0, NULL);

	for(map <string, vector < pair <int, pair <int, int> > > >::iterator titer = fstpostings.begin(); titer != fstpostings.end(); titer++) {
		string thisid = titer->first;
		if(theargs.V) cout << "Computing for tuple dataid " << thisid << endl;
		PGresult* resgph;
		const char *paramValuesgph[1];
		paramValuesgph[0] = (const char*) thisid.c_str();
		resgph = PQexecPrepared(conn, "getgph", 1, paramValuesgph, NULL, NULL, 1);
		PQexecPrepared(conn, "begintr", 0, NULL, NULL, NULL, 1);
		Oid objid = unpack_int32(PQgetvalue(resgph, 0, 0));
		int objfd = lo_open(conn, objid, INV_READ);
		lo_lseek(conn, objfd, 0, SEEK_END);
		int bsize = lo_tell(conn, objfd);
		lo_lseek(conn, objfd, 0, SEEK_SET);
		char buffer[bsize];
		memset(buffer, bsize, '\0');
		int retlo = lo_read(conn, objfd, buffer, bsize);
		if(retlo < 0) {
			if(theargs.V) cout << "Error: Warning to read graph blob at dataid " << thisid << endl;
			continue;
		}
		lo_close(conn, objfd);
		PQexecPrepared(conn, "endtr", 0, NULL, NULL, NULL, 1);

		structureInfo stacinfo;	
		graphread(&stacinfo, buffer);
		PQclear(resgph);
	
		map<int, vector<int> > stacgraphproj;				//the projected sublkgph
		map<int, double> intersblkprob;							//reachprob for inter sblks (non-MS) such that we ignore paths thro any of its ancestors (disjreachprob for case 2)
		int regularcaseflag;												//whether or not the query procng is regular as in filescan -> and we just mult the uniq startblk's reachprob
		regularcaseflag = getProjection(theargs.L, stacinfo, titer->second, &stacgraphproj, &intersblkprob);
		stacinfo.stacgraph = stacgraphproj;
		map <int, vector <int> > stacgraphrev;	
		reverseGraph(&stacgraphproj, &stacgraphrev);
		
		//read these blocks alone from the database and construct the prob matrices for them
		PGresult* resti;
		string subcomdstr = "select blocknum, path, pathwt from " + stacgmtablename + " where dataid = " + thisid + " and (";
		for(map <int, vector <int> >::iterator iter2 = stacgraphproj.begin(); iter2 != stacgraphproj.end(); iter2++) {
			stringstream iterstr;
			iterstr << iter2->first;
			subcomdstr += "blocknum = " + iterstr.str() + " or ";	
		}
		subcomdstr += " false);";	//end with false for filler
		resti = PQexecParams(conn, (char*) subcomdstr.c_str(), 0, NULL, NULL, NULL, NULL, 1);	
		if (PQresultStatus(resti) != PGRES_TUPLES_OK) {
			cout << "Error: Select tuples failed: " << PQerrorMessage(conn);
			PQclear(resti);
			PQfinish(conn);
			return 1;
		}
	
		int numtti = PQntuples(resti);
		if(numtti <= 0) {
			if(theargs.V) cout << "Warning: No tuples were read for dataid " << thisid << "\n";
			continue;
		}
			
		int numblks = stacinfo.stacgraph.size();		//use newblkno when accessing these mem arrays; cannot assume blknos are from 0 to numblks-1; need a new mapping
		map <int, int> newblkno;	//maps from orig blkno to new blkno in subgph (0 to numblks-1)
		int newnum = 0;
		for(map<int, vector<int> >::iterator git = stacinfo.stacgraph.begin(); git != stacinfo.stacgraph.end(); git++)
			newblkno[git->first] = newnum++;
		double* transprob = (double *) malloc (sizeof(double) * numblks * NUMST * NUMST);	//allocate array of matrices for each block in gph
		memset(transprob, '\0', sizeof(double) * numblks * NUMST * NUMST);
		for(int x=0; x<numtti; x++) {
			char* bnox = PQgetvalue(resti, x, 0);
			char* bpath = PQgetvalue(resti, x, 1);
			char* bwt = PQgetvalue(resti, x, 2);
			int nm = unpack_int32(bnox);
			nm = newblkno[nm];		//convert from orig blkno to new blkno
			string pathx = string(bpath);
			double wt = unpack_double(bwt);
			for(int inp=0; inp<NUMST; inp++) {
				int linenxtst = dfacheck(pathx, dfaquery, inp , NUMST, startstate, finalstate);
				transprob [nm * NUMST * NUMST + inp * NUMST + linenxtst] += exp(-1.0*wt);
			}
		}
		PQclear(resti);

		int* memo = (int *) malloc (sizeof(int) * numblks);
		memset(memo, '\0', sizeof(int) * numblks);
		vector <int> revstartblks;
		for(map <int, vector <int> >::iterator iter = stacinfo.stacgraph.begin(); iter != stacinfo.stacgraph.end(); iter++)
			if(iter->second.size() == 0)		//outdeg is 0 means it is a revstartblk
				revstartblks.push_back(iter->first);

		double* probmatr = (double *) malloc (sizeof(double) * NUMST * NUMST);	//this is the final matrix that adds all revstartblks result matrices
		memset(probmatr, '\0', sizeof(double) * NUMST * NUMST);
		double* scratches = (double *) malloc (sizeof(double) * 2 * NUMST * NUMST);	//scratches for addn and mult of matrices; memset done in function
		for(vector <int>::iterator siter = revstartblks.begin(); siter != revstartblks.end(); siter++) {
			if(regularcaseflag == 1)
				dynmatrdfs (&stacgraphrev, &stacinfo.blkreachprobs, transprob, scratches, memo, NUMST, *siter, &newblkno);
			else
				dynmatrdfsinter (&stacgraphrev, &stacinfo.blkreachprobs, transprob, scratches, memo, NUMST, *siter, &newblkno, &intersblkprob);
			int siterint = newblkno[*siter];
			incrsummatrs (probmatr, transprob + ((siterint) * NUMST * NUMST), NUMST);	
		}	

		double acceptprob = probmatr[startstate * NUMST + finalstate];
		free(transprob);
		free(memo);
		free(scratches);
		free(probmatr);
		restuple thistuple;
		if(acceptprob > 0) {	//only consider those with matchprob > 0
			thistuple.matchwt = -1.0*log(acceptprob);
			string cmdstringaux = qstructs.constructCmdId(thisid);
			PGresult* auxres = PQexec(conn, cmdstringaux.c_str());
			for(int colv=0; colv<numac; colv++)
				thistuple.precolvals.push_back(PQgetvalue(auxres, 0, 2+colv));
			PQclear(auxres);
			results.push_back(thistuple);
		}		
	}//end for on fsts
	
	PQclear(res);
	PQfinish(conn);
	
	sortResults(&results);
	printResults(results, qstructs.auxcolnames, qstructs.Ktop);

return 0;
} 
